use std::fmt;

use serde::Serialize;

#[derive(Debug, PartialEq, Clone, Serialize)]
pub struct Person<'a> {
    first_name: Option<String>,
    last_name: String,
    email: String,
    #[serde(skip_serializing)]
    org: Option<&'a Organisation<'a>>,
}

#[derive(Debug, PartialEq, Clone, Serialize)]
pub struct Organisation<'a> {
    name: String,
    email: String,
    #[serde(skip_serializing)]
    members: Vec<&'a Person<'a>>,
}

#[derive(Debug, PartialEq, Clone, Serialize)]
pub enum Contact<'a> {
    Person(Person<'a>),
    Organisation(Organisation<'a>),
}

impl<'a> Person<'a> {
    pub fn new(last_name: String, email: String) -> Self {
        Person {
            first_name: None,
            last_name,
            email,
            org: None,
        }
    }

    // Builder Pattern
    pub fn with_first_name(mut self, first_name: String) -> Self {
        self.first_name = Some(first_name);
        self
    }

    pub fn with_org(mut self, org: &'a Organisation) -> Self {
        self.org = Some(org);
        self
    }

    /// Retrieve the mail address of the person
    ///
    /// # Example
    ///
    /// ```
    /// use kontakte::types::Person;
    ///
    /// let tux = Person::new("Pinguin".to_string(), "tux@example.com".to_string());
    /// assert_eq!(tux.email(), "tux@example.com");
    /// ```
    pub fn email(&self) -> &str {
        &self.email
    }
}

impl fmt::Display for Person<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.first_name {
            Some(ref first_name) => write!(f, "{} {} <{}>", first_name, self.last_name, self.email),
            None => write!(f, "{} <{}>", self.last_name, self.email),
        }

        // Alternative:
        // format!("{} {} <{}>", self.first_name.as_ref().unwrap_or(&"".to_string()), self.last_name, self.email)
    }
}

impl<'a> Organisation<'a> {
    pub fn new(name: String, email: String) -> Self {
        Organisation {
            name,
            email,
            members: Vec::new(),
        }
    }

    pub fn add_member(&mut self, person: &'a Person) {
        self.members.push(person)
    }
}

impl fmt::Display for Organisation<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{} <{}>", self.name, self.email)
    }
}

impl fmt::Display for Contact<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Person(p) => p.fmt(f),
            Self::Organisation(o) => o.fmt(f),
        }
    }
}

// Contact.from(person)
// let contact: Contact = Person.into()
impl<'a> From<Person<'a>> for Contact<'a> {
    fn from(value: Person<'a>) -> Self {
        Self::Person(value)
    }
}

impl<'a> From<Organisation<'a>> for Contact<'a> {
    fn from(value: Organisation<'a>) -> Self {
        Self::Organisation(value)
    }
}
